from ... import exceptions


class WalletRESTException(exceptions.BackendException, exceptions.WalletException):
    def __init__(self, *args, **kwargs):
        self.result = kwargs.pop("result", None)
        super(WalletRESTException, self).__init__(*args, **kwargs)


class BadRequest(WalletRESTException):
    """Raised when the underlying REST API returns HTTP code 400."""

    pass


class NotFound(WalletRESTException):
    """Raised when the underlying REST API returns HTTP code 404."""

    pass


class RESTServerError(WalletRESTException):
    """
    Raised when the underlying REST API returns HTTP code 403 or 500 and the error cannot be
    handled.
    """

    pass


class CreatedInvalidTransaction(WalletRESTException, exceptions.TransactionException):
    pass


class NotSupported(WalletRESTException):
    """
    Raised when wallet doesn't provide the requested feature.
    """

    pass



import collections
import operator
import re
import warnings

from .address import Address, address
from .metadata import Metadata
from .numbers import as_ada

__all__ = ("Transaction", "Input", "Output", "validate_txid")


class Transaction(object):
    """
    Represents a Cardano transaction.

    :param txid:            the ID of the transaction
    :param fee:             fee amount in ADA
    :param inputs:          a sequence of :class:`Input` objects
    :param outputs:         a sequence of :class:`Output` objects
    :param local_inputs:    a sequence of :class:`Input` objects that originate from local wallet
    :param local_outputs:   a sequence of :class:`Output` objects that are destined to local wallet
    :param withdrawals:     a sequence of (:class:`Decimal`, :class:`str`) pairs of amounts and
                            stake addresses
    :param metadata:        an instance of :class:`Metadata <cardano.metadata.Metadata>`
    """

    txid = None
    fee = None
    inputs = None
    outputs = None
    local_inputs = None
    local_outputs = None
    withdrawals = None
    inserted_at = None
    expires_at = None
    pending_since = None
    status = None
    metadata = None

    def __init__(self, txid=None, **kwargs):
        self.txid = txid or self.txid
        validate_txid(self.txid)
        fee = kwargs.pop("fee", None)
        self.fee = fee if fee is not None else self.fee
        self.inputs = kwargs.pop("inputs", []) or (
            self.inputs if self.inputs is not None else []
        )
        self.outputs = kwargs.pop("outputs", []) or (
            self.outputs if self.outputs is not None else []
        )
        self.local_inputs = kwargs.pop("local_inputs", []) or (
            self.local_inputs if self.local_inputs is not None else []
        )
        self.local_outputs = kwargs.pop("local_outputs", []) or (
            self.local_outputs if self.local_outputs is not None else []
        )
        self.withdrawals = kwargs.pop("withdrawals", []) or (
            self.withdrawals if self.withdrawals is not None else []
        )
        self.inserted_at = kwargs.pop("inserted_at", None) or self.inserted_at
        self.expires_at = kwargs.pop("expires_at", None) or self.expires_at
        self.pending_since = kwargs.pop("pending_since", None) or self.pending_since
        self.status = kwargs.pop("status", None) or self.status
        self.metadata = kwargs.pop("metadata", Metadata()) or (
            self.metadata if self.metadata is not None else Metadata()
        )

    def __repr__(self):
        return "<Cardano tx: {:s}>".format(self.txid)

    def __format__(self, spec):
        return format(str(self), spec)

    @property
    def local_inputs_sum(self):
        return as_ada(sum(map(operator.attrgetter("amount"), self.local_inputs)))

    @property
    def local_outputs_sum(self):
        return as_ada(sum(map(operator.attrgetter("amount"), self.local_outputs)))

    @property
    def amount_in(self):
        return as_ada(max(0, self.local_outputs_sum - self.local_inputs_sum))

    @property
    def amount_out(self):
        return as_ada(
            max(
                0,
                self.local_inputs_sum - self.local_outputs_sum - self.fee
                if self.fee is not None
                else 0,
            )
        )

    #    @property
    #    def assets(self):
    #        _assets = collections.defaultdict(int)
    #        print("i")
    #        for inp in self.inputs:
    #            print(inp.assets)
    #            for aid, aqty in inp.assets:
    #                _assets[aid] += -aqty if inp in self.local_inputs else aqty
    #                print(dict(_assets))
    #        print("o")
    #        for outp in self.outputs:
    #            print(outp.assets)
    #            for aid, aqty in outp.assets:
    #                print(outp in self.local_outputs)
    #                _assets[aid] += aqty if outp in self.local_outputs else -aqty
    #                print(dict(_assets))
    #        return dict(_assets)

    def hash(self):
        return self.txid


class IOBase(object):
    def __init__(self, address=None, amount=None, assets=None):
        self.address = None if address is None else Address(address)
        self.amount = amount
        self.assets = assets or []


class Input(IOBase):
    """
    Represents a :class:`Transaction` input.

    :param iid:     the input ID
    :type iid:      :class:`str` hex
    :param address: the origin address
    :type address:  :class:`cardano.address.Address`
    :param amount:  the amount in ADA
    :type amount:   :class:`Decimal`
    :param assets:  a sequence of :class:`AssetID <cardano.simpletypes.AssetID>` quantity pairs
    :type assets:   :class:`list`
    """

    def __init__(self, iid=None, address=None, amount=None, assets=None):
        super(Input, self).__init__(address=address, amount=amount, assets=assets)
        self.iid = iid


class Output(IOBase):
    """
    Represents a :class:`Transaction` output.

    :param address: the destination address
    :type address:  :class:`cardano.address.Address`
    :param amount:  the amount in ADA
    :type amount:   :class:`Decimal`
    :param assets:  a sequence of :class:`AssetID <cardano.simpletypes.AssetID>` quantity pairs
    :type assets:   :class:`list`
    """

    pass


def validate_txid(txid):
    if not bool(re.compile("^[0-9a-f]{64}$").match(txid)):
        raise ValueError(
            "Transaction ID must be a 64-character hexadecimal string, not "
            "'{}'".format(txid)
        )
    return txid


class TransactionManager(object):
    wid = None
    backend = None

    def __init__(self, wid, backend):
        self.wid = wid
        self.backend = backend

    def __call__(self, **filterparams):
        filter_ = TxFilter(**filterparams)
        return filter_.filter(self.backend.transactions(self.wid))


class _ByHeight(object):
    """A helper class used as key in sorting of payments by height.
    Mempool goes on top, blockchain payments are ordered with descending block numbers.

    **WARNING:** Integer sorting is reversed here.
    """

    def __init__(self, tx):
        self.tx = tx

    def _cmp(self, other):
        sh = self.tx.inserted_at
        oh = other.tx.inserted_at
        if sh is oh is None:
            return 0
        if sh is None:
            return 1
        if oh is None:
            return -1
        return (sh.absolute_slot > oh.absolute_slot) - (
            sh.absolute_slot < oh.absolute_slot
        )

    def __lt__(self, other):
        return self._cmp(other) > 0

    def __le__(self, other):
        return self._cmp(other) >= 0

    def __eq__(self, other):
        return self._cmp(other) == 0

    def __ge__(self, other):
        return self._cmp(other) <= 0

    def __gt__(self, other):
        return self._cmp(other) < 0

    def __ne__(self, other):
        return self._cmp(other) != 0


class TxFilter(object):
    #
    # Available filters:
    # - txid
    # - src_addr
    # - dest_addr
    # - min_epoch
    # - max_epoch
    # - min_slot
    # - max_slot
    # - min_absolute_slot
    # - max_absolute_slot
    # - min_height
    # - max_height
    # - confirmed
    # - unconfirmed
    #
    def __init__(self, **filterparams):
        self.min_epoch = filterparams.pop("min_epoch", None)
        self.max_epoch = filterparams.pop("max_epoch", None)
        self.min_slot = filterparams.pop("min_slot", None)
        self.max_slot = filterparams.pop("max_slot", None)
        self.min_absolute_slot = filterparams.pop("min_absolute_slot", None)
        self.max_absolute_slot = filterparams.pop("max_absolute_slot", None)
        self.min_height = filterparams.pop("min_height", None)
        self.max_height = filterparams.pop("max_height", None)
        self.unconfirmed = filterparams.pop("unconfirmed", False)
        self.confirmed = filterparams.pop("confirmed", True)
        _txid = filterparams.pop("txid", None)
        _src_addr = filterparams.pop("src_addr", None)
        _dest_addr = filterparams.pop("dest_addr", None)
        if len(filterparams) > 0:
            raise ValueError(
                "Excessive arguments for payment query: {}".format(filterparams)
            )
        self._asks_chain_position = any(
            map(
                lambda x: x is not None,
                (
                    self.min_epoch,
                    self.max_epoch,
                    self.min_slot,
                    self.max_slot,
                    self.min_absolute_slot,
                    self.max_absolute_slot,
                    self.min_height,
                    self.max_height,
                ),
            )
        )
        if self.unconfirmed and self._asks_chain_position:
            warnings.warn(
                "Blockchain position filtering ({max,min}_{epoch,slot,block}) has been "
                "requested while also asking for transactions not in ledger. "
                "These are mutually exclusive. "
                "As mempool transactions have no height at all, they will be excluded "
                "from the result.",
                RuntimeWarning,
            )
        self.src_addrs = self._get_addrset(_src_addr)
        self.dest_addrs = self._get_addrset(_dest_addr)
        if _txid is None:
            self.txids = []
        else:
            if isinstance(_txid, (bytes, str)):
                txids = [_txid]
            else:
                iter(_txid)
                txids = _txid
            self.txids = list(map(validate_txid, txids))

    def _get_addrset(self, addr):
        if addr is None:
            return set()
        else:
            if isinstance(addr, (str, bytes)):
                addrs = [addr]
            else:
                try:
                    iter(addr)
                    addrs = addr
                except TypeError:
                    addrs = [addr]
            return set(map(address, addrs))

    def check(self, tx):
        assert (tx.status == "in_ledger" and tx.inserted_at is not None) or (
            tx.status != "in_ledger" and tx.inserted_at is None
        )
        ht = tx.inserted_at
        if ht is None:
            if not self.unconfirmed:
                return False
            if self._asks_chain_position:
                # mempool txns are filtered out if any height range check is present
                return False
        else:
            if not self.confirmed:
                return False
            if self.min_epoch is not None and ht.epoch < self.min_epoch:
                return False
            if self.max_epoch is not None and ht.epoch > self.max_epoch:
                return False
            if self.min_slot is not None and ht.slot < self.min_slot:
                return False
            if self.max_slot is not None and ht.slot > self.max_slot:
                return False
            if (
                self.min_absolute_slot is not None
                and ht.absolute_slot < self.min_absolute_slot
            ):
                return False
            if (
                self.max_absolute_slot is not None
                and ht.absolute_slot > self.max_absolute_slot
            ):
                return False
            if self.min_height is not None and ht.height < self.min_height:
                return False
            if self.max_height is not None and ht.height > self.max_height:
                return False
        if self.txids and tx.txid not in self.txids:
            return False
        srcs = set(filter(None, map(operator.attrgetter("address"), tx.inputs)))
        dests = set(filter(None, map(operator.attrgetter("address"), tx.inputs)))
        if self.src_addrs and not self.src_addrs.intersection(srcs):
            return False
        if self.dest_addrs and not self.dest_addrs.intersection(dests):
            return False
        return True

    def filter(self, txns):
        return sorted(filter(self.check, txns), key=_ByHeight)



from binascii import hexlify, unhexlify
from decimal import Decimal
import json
import logging
import operator
import requests
import urllib

from ... import exceptions as main_exceptions
from ...metadata import Metadata
from ...numbers import from_lovelaces, to_lovelaces
from ...simpletypes import (
    AssetID,
    Balance,
    StakePoolInfo,
    StakePoolStatus,
    StakeRewardMetrics,
    StakingStatus,
)
from ...transaction import Transaction
from . import exceptions
from . import serializers

_log = logging.getLogger(__name__)


class JSONWithDecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Decimal):
            return str(o)
        return super(JSONWithDecimalEncoder, self).default(o)


class WalletREST(object):
    base_url = None
    timeout = 10

    ERR2EXCEPTION = {
        403: {
            "pool_already_joined": main_exceptions.PoolAlreadyJoined,
            "non_null_rewards": main_exceptions.NonNullRewards,
            "cannot_cover_fee": main_exceptions.CannotCoverFee,
        },
        500: {
            "created_invalid_transaction": exceptions.CreatedInvalidTransaction,
        },
    }

    def __init__(self, protocol="http", host="localhost", port=8090):
        self.base_url = "{protocol}://{host}:{port}/v2/".format(
            protocol=protocol, host=host, port=port
        )
        _log.debug("WalletREST backend url: {:s}".format(self.base_url))

    def raw_request(self, method, path, params=None):
        url = "".join([self.base_url, path])
        hdr = {"Content-Type": "application/json"}
        params = params or {}
        _log.debug(
            u"{method} {url}\nParams:\n{params}".format(
                method=method,
                url=url,
                params=json.dumps(params, indent=2, sort_keys=True),
            )
        )
        rsp = getattr(requests, method.lower())(
            url, headers=hdr, data=json.dumps(params), timeout=self.timeout
        )
        if rsp.status_code != 204:  # if content exists
            result = rsp.json(parse_float=Decimal)
            _ppresult = json.dumps(
                rsp.json(), cls=JSONWithDecimalEncoder, indent=2, sort_keys=True
            )
            _log.debug(u"Result:\n{result}".format(result=_ppresult))
        else:
            result = None
            _log.debug(u"No result (HTTP 204)")
        if rsp.status_code == 400:
            raise exceptions.BadRequest(result["message"], result=result)
        if rsp.status_code == 403:
            try:
                raise self.ERR2EXCEPTION[rsp.status_code][result["code"]](
                    result["message"]
                )
            except KeyError:
                pass
            raise exceptions.RESTServerError(result.get("message", "* NO MESSAGE *"))
        if rsp.status_code == 404:
            raise exceptions.NotFound(result["message"], result=result)
        if rsp.status_code == 500:
            try:
                raise self.ERR2EXCEPTION[rsp.status_code][result["code"]](
                    result["message"], result=result
                )
            except KeyError:
                pass
            raise exceptions.RESTServerError(
                result.get("message", "* NO MESSAGE *"), result=result
            )
        return result

    def wallet_ids(self):
        return map(operator.itemgetter("id"), self.raw_request("GET", "wallets"))

    def wallet_exists(self, wid):
        try:
            self.raw_request("GET", "wallets/{:s}".format(wid))
        except exceptions.NotFound:
            return False
        return True

    def create_wallet(self, name, mnemonic, passphrase, mnemonic_2f=None):
        data = {
            "name": name,
            "mnemonic_sentence": mnemonic.split(),
            "passphrase": passphrase,
            "address_pool_gap": 20,
        }
        if mnemonic_2f:
            data["mnemonic_second_factor"] = mnemonic_2f.split()
        wdata = self.raw_request("POST", "wallets", data)
        return wdata["id"]

    def delete_wallet(self, wid):
        try:
            self.raw_request("DELETE", "wallets/{:s}".format(wid))
        except exceptions.NotFound:
            return False
        return True

    def sync_progress(self, wid):
        wdata = self.raw_request("GET", "wallets/{:s}".format(wid))
        return (
            1.0
            if wdata["state"]["status"] == "ready"
            else wdata["state"]["progress"]["quantity"] / Decimal(100)
        )

    def balance(self, wid):
        bdata = self.raw_request("GET", "wallets/{:s}".format(wid))["balance"]
        return Balance(
            from_lovelaces(bdata["total"]["quantity"]),
            from_lovelaces(bdata["available"]["quantity"]),
            from_lovelaces(bdata["reward"]["quantity"]),
        )

    def asset_balances(self, wid):
        data = self.raw_request("GET", "wallets/{:s}".format(wid))
        try:
            bdata = data["assets"]
        except KeyError:
            raise exceptions.NotSupported("Extra assets are not supported")
        assets = {}
        for ast in bdata["total"]:
            aid = AssetID(ast["asset_name"], ast["policy_id"])
            assets[aid] = assets[aid] if aid in assets else {}
            assets[aid]["total"] = ast["quantity"]
        for ast in bdata["available"]:
            aid = AssetID(ast["asset_name"], ast["policy_id"])
            assets[aid] = assets[aid] if aid in assets else {}
            assets[aid]["available"] = ast["quantity"]
        return {
            aid: Balance(bal["total"], bal["available"], None)
            for aid, bal in assets.items()
        }

    def addresses(self, wid):
        adata = self.raw_request("GET", "wallets/{:s}/addresses".format(wid))
        return [(ad["id"], True if ad["state"] == "used" else False) for ad in adata]

    def _addresses_set(self, wid):
        return set(map(operator.itemgetter(0), self.addresses(wid)))

    def _txdata2tx(self, txd, addresses=None):
        inputs = (
            [serializers.get_input(inp) for inp in txd["inputs"]]
            if "inputs" in txd
            else []
        )
        outputs = (
            [serializers.get_output(outp) for outp in txd["outputs"]]
            if "outputs" in txd
            else []
        )
        local_inputs = set()
        local_outputs = set()
        amount = Decimal(0)
        if addresses:
            for out in outputs:
                if out.address and out.address in addresses:
                    local_outputs.add(out)
                    amount += out.amount
            for inp in inputs:
                if inp.address and inp.address in addresses:
                    local_inputs.add(inp)
        fee = serializers.get_amount(txd["fee"])
        metadata = (
            Metadata.deserialize(txd["metadata"])
            if txd.get("metadata", None) is not None
            else None
        )
        return Transaction(
            txid=txd["id"],
            amount=amount,
            fee=fee,
            inserted_at=serializers.get_block_position(txd["inserted_at"])
            if "inserted_at" in txd
            else None,
            expires_at=serializers.get_block_position(txd["expires_at"])
            if "expires_at" in txd
            else None,
            pending_since=serializers.get_block_position(txd["pending_since"])
            if "pending_since" in txd
            else None,
            inputs=inputs,
            outputs=outputs,
            local_inputs=local_inputs,
            local_outputs=local_outputs,
            withdrawals=[
                (serializers.get_amount(w["amount"]), w["stake_address"])
                for w in txd.get("withdrawals", [])
            ],
            status=txd["status"],
            metadata=metadata,
        )

    def transactions(self, wid):
        data = {"order": "ascending"}
        return [
            self._txdata2tx(txd, addresses=self._addresses_set(wid))
            for txd in self.raw_request(
                "GET", "wallets/{:s}/transactions".format(wid), data
            )
        ]

    def transfer(self, wid, destinations, metadata, allow_withdrawal, ttl, passphrase):
        data = {
            "passphrase": passphrase,
            "withdrawal": "self" if allow_withdrawal else None,
            "payments": [
                {
                    "address": str(address),
                    "amount": serializers.store_amount(amount),
                    "assets": [
                        {
                            "policy_id": asset.policy_id,
                            "asset_name": asset.asset_name,
                            "quantity": asset_amount,
                        }
                        for (asset, asset_amount) in assets
                    ],
                }
                for (address, amount, assets) in destinations
            ],
        }
        if metadata is not None:
            if not isinstance(metadata, Metadata):
                metadata = Metadata(metadata.items())
            data["metadata"] = metadata.serialize()
        if ttl is not None:
            data["time_to_live"] = serializers.store_interval(ttl)
        # NOTE: the order of the following two requests is important
        txd = self.raw_request("POST", "wallets/{:s}/transactions".format(wid), data)
        return self._txdata2tx(txd, addresses=self._addresses_set(wid))

    def estimate_fee(self, wid, destinations, metadata):
        data = {
            "payments": [
                {
                    "address": str(address),
                    "amount": serializers.store_amount(amount),
                }
                for (address, amount) in destinations
            ],
        }
        if metadata is not None:
            if not isinstance(metadata, Metadata):
                metadata = Metadata(metadata.items())
            data["metadata"] = metadata.serialize()
        feedata = self.raw_request(
            "POST", "wallets/{:s}/payment-fees".format(wid), data
        )
        return (
            serializers.get_amount(feedata["estimated_min"]),
            serializers.get_amount(feedata["estimated_max"]),
        )

    def _stakepoolinfo(self, pooldata, stake):
        retirement = (
            serializers.get_epoch(pooldata["retirement"])
            if "retirement" in pooldata and "epoch_number" in pooldata["retirement"]
            else None
        )
        status = (
            StakePoolStatus.DELISTED
            if "flags" in pooldata and "delisted" in pooldata["flags"]
            else StakePoolStatus.RETIRING
            if retirement
            else StakePoolStatus.ACTIVE
        )
        if "metadata" in pooldata:
            ticker = pooldata["metadata"]["ticker"]
            name = pooldata["metadata"]["name"]
            description = pooldata["metadata"]["description"]
            homepage = pooldata["metadata"]["homepage"]
        else:
            ticker = name = description = homepage = None
        rewards = StakeRewardMetrics(
            serializers.get_amount(pooldata["metrics"]["non_myopic_member_rewards"]),
            stake,
        )
        return StakePoolInfo(
            pooldata["id"],
            status,
            ticker,
            name,
            description,
            homepage,
            rewards,
            serializers.get_amount(pooldata["cost"]),
            serializers.get_percent(pooldata["margin"]),
            serializers.get_amount(pooldata["pledge"]),
            serializers.get_percent(pooldata["metrics"]["relative_stake"]),
            pooldata["metrics"]["saturation"],
            pooldata["metrics"]["produced_blocks"]["quantity"],
            retirement,
        )

    def stake_pools(self, wid, stake):
        urldata = {"stake": to_lovelaces(stake)}
        poolsdata = self.raw_request(
            "GET", "?".join(("stake-pools", urllib.parse.urlencode(urldata)))
        )
        return [self._stakepoolinfo(pool, stake) for pool in poolsdata]

    def _stakingstatus(self, data):
        return StakingStatus(
            serializers.get_stakingstatus(data["status"]),
            data["target"] if "target" in data else None,
            serializers.get_epoch(data["changes_at"]) if "changes_at" in data else None,
        )

    def staking_status(self, wid):
        sdata = self.raw_request("GET", "wallets/{:s}".format(wid))["delegation"]
        active = sdata["active"]
        return (
            self._stakingstatus(active),
            [self._stakingstatus(ss) for ss in sdata["next"]],
        )

    def stake(self, wid, pool_id, passphrase):
        txdata = self.raw_request(
            "PUT",
            "stake-pools/{:s}/wallets/{:s}".format(pool_id, wid),
            {"passphrase": passphrase},
        )
        return self._txdata2tx(txdata, addresses=self._addresses_set(wid))

    def unstake(self, wid, passphrase):
        txdata = self.raw_request(
            "DELETE",
            "stake-pools/*/wallets/{:s}".format(wid),
            {"passphrase": passphrase},
        )
        return self._txdata2tx(txdata, addresses=self._addresses_set(wid))

    def utxo_stats(self, wid):
        sdata = self.raw_request(
            "GET",
            "wallets/{:s}/statistics/utxos".format(wid),
            {},
        )
        return (
            serializers.get_amount(sdata["total"]),
            {
                from_lovelaces(int(lvl)): num
                for (lvl, num) in sorted(
                    sdata["distribution"].items(), key=operator.itemgetter(0)
                )
            },
            sdata["scale"],
        )



import binascii
import collections
import enum


Balance = collections.namedtuple("Balance", ["total", "available", "reward"])
Balance.__doc__ = "Represents a balance of asset, including total, principal and reward"
Balance.total.__doc__ = "The total balance"
Balance.available.__doc__ = "The principal, i.e. the total minus staking rewards"
Balance.reward.__doc__ = "The staking rewards (interest)"


BlockPosition = collections.namedtuple(
    "BlockPosition", ["epoch", "slot", "absolute_slot", "height"]
)
BlockPosition.__doc__ = "Represents block's position within the blockchain"
BlockPosition.epoch.__doc__ = "Epoch number"
BlockPosition.slot.__doc__ = "Slot number"
BlockPosition.absolute_slot.__doc__ = "Absolute slot number"
BlockPosition.height.__doc__ = "Block number (height of the chain) [optional]"


Epoch = collections.namedtuple("Epoch", ["number", "starts"])


class AssetID(object):
    """
    Represents the ID of a native Cardano asset. It consists of asset name and policy ID.
    It renders as string representation of ``asset_name:policy_id``.

    The ``asset_name`` is always kept encoded as hexadecimal string and must be passed
    to the constructor as such.

    The ``.name_bytes`` property is a :class:`bytes` decoded representation of the hex.
    Because Cardano allows full ASCII set to be used in asset names, some of them are not
    safe to be displayed directly.
    """

    asset_name = ""
    policy_id = None
    name_bytes = None

    def __init__(self, asset_name, policy_id):
        asset_name = asset_name if asset_name is not None else self.asset_name
        policy_id = policy_id or self.policy_id
        # binascii.hexlify() returns bytes() for some unknown reason. We may expect them to be
        # passed here:
        if isinstance(asset_name, bytes):
            self.name_bytes = binascii.unhexlify(asset_name)
            self.asset_name = asset_name.decode()
        elif isinstance(asset_name, str):
            self.name_bytes = binascii.unhexlify(asset_name.encode())
            self.asset_name = asset_name
        else:
            raise ValueError(
                "The asset_name is neither str or bytes but {}".format(
                    type(self.asset_name)
                )
            )
        self.policy_id = policy_id

    def __repr__(self):
        return "{:s}:{:s}".format(self.asset_name, self.policy_id)

    def __eq__(self, other):
        if isinstance(other, AssetID):
            return str(self) == str(other)
        elif isinstance(other, str):
            return str(self) == other
        elif isinstance(other, bytes):
            return str(self).encode() == other
        return super(AssetID, self).__eq__(other)

    def __hash__(self):
        return hash(str(self))


StakePoolStatus = enum.Enum("StakePoolStatus", "ACTIVE RETIRING DELISTED")
StakePoolStatus.__doc__ = "Represents stake pool status"

StakeRewardMetrics = collections.namedtuple(
    "StakeRewardMetrics",
    [
        "expected",
        "stake",
    ],
)
StakeRewardMetrics.__doc__ = "Represents stake pool reward metrics"
StakeRewardMetrics.expected.__doc__ = "Expected rewards at the end of an epoch, in ADA"
StakeRewardMetrics.stake.__doc__ = (
    "Staked amount against which rewards were calculated, in ADA"
)

StakePoolInfo = collections.namedtuple(
    "StakePoolInfo",
    [
        "id",
        "status",
        "ticker",
        "name",
        "description",
        "homepage",
        "rewards",
        "cost",
        "margin",
        "pledge",
        "relative_stake",
        "saturation",
        "produced_blocks",
        "retirement",
    ],
)
StakePoolInfo.__doc__ = "Stores stake pool data"
StakePoolInfo.id.__doc__ = "Unique ID"
StakePoolInfo.status.__doc__ = "Status, one of :class:`StakePoolStatus` enum"
StakePoolInfo.ticker.__doc__ = "3-5 chars long ticker"
StakePoolInfo.name.__doc__ = "Name"
StakePoolInfo.description.__doc__ = "Description"
StakePoolInfo.homepage.__doc__ = "Homepage URL"
StakePoolInfo.cost.__doc__ = "Fixed pool running cost in ADA"
StakePoolInfo.margin.__doc__ = "Operator's margin on the total reward before splitting it among stakeholders (as :class:`Decimal` fraction)"
StakePoolInfo.pledge.__doc__ = "Minimal stake amount that the pool is willing to honor"
StakePoolInfo.relative_stake.__doc__ = "The live pool stake relative to the total stake"
StakePoolInfo.saturation.__doc__ = (
    "Saturation-level of the pool based on the desired number "
    "of pools aimed by the network. A value above 1 indicates that the pool is saturated."
)
StakePoolInfo.produced_blocks.__doc__ = (
    "Number of blocks produced by a given stake pool in its lifetime."
)
StakePoolInfo.retirement.__doc__ = "The :class:`Epoch` in which the pool retires"

StakingStatus = collections.namedtuple(
    "StakingStatus",
    [
        "delegating",
        "target_id",
        "changes_at",
    ],
)
StakingStatus.__doc__ = "Wallet's staking status"
StakingStatus.delegating.__doc__ = "Whether the wallet is actively delegating"
StakingStatus.target_id.__doc__ = "The ID of the pool the wallet is delegating to"
StakingStatus.changes_at.__doc__ = ":class:`Epoch` since which the change comes live"



from .consts import Era


def address(addr, wallet=None):
    if isinstance(addr, Address):
        return addr  # already instatinated and should be of proper class
    elif isinstance(addr, (bytes, bytearray)):
        addr = addr.decode()
    elif not isinstance(addr, str):
        raise TypeError(
            "address() argument must be str, bytes, bytearray or Address instance"
        )
    # validation
    if (
        addr.startswith("addr1")
        or addr.startswith("addr_test1")
        or addr.startswith("stake1")
        or addr.startswith("stake_test1")
    ):
        AddressClass = ShelleyAddress
    elif addr.startswith("DdzFF"):
        AddressClass = ByronAddress
    elif addr.startswith("Ae2"):
        AddressClass = IcarusAddress
    else:
        raise ValueError("String {} is not a valid Cardano address".format(addr))
    return AddressClass(addr, wallet=wallet)


class Address(object):
    """
    Cardano base address class. Does no validation, it is up to child classes.

    Compares with ``str`` and ``bytes`` objects.

    :param addr:    the address as ``str`` or ``bytes`` or ``Address``
    :param wallet:  the ``Wallet`` object if address belongs to
    """

    _address = ""
    wallet = None

    def __init__(self, addr, wallet=None):
        self._address = addr
        self.wallet = wallet or self.wallet

    def __repr__(self):
        return str(self._address)

    def __eq__(self, other):
        if isinstance(other, Address):
            return str(self) == str(other)
        elif isinstance(other, str):
            return str(self) == other
        elif isinstance(other, bytes):
            return str(self).encode() == other
        return super(Address, self).__eq__(other)

    def __hash__(self):
        return hash(str(self))

    def __format__(self, spec):
        return format(str(self), spec)


class ByronAddress(Address):
    era = Era.BYRON


class IcarusAddress(ByronAddress):
    pass


class ShelleyAddress(Address):
    era = Era.SHELLEY
